package src;

import java.io.*;
import java.net.*;
import java.util.Vector;
import java.util.HashMap;
import java.util.concurrent.ConcurrentLinkedQueue;

public class MbsModule
{
	private QueuePool qpool;
	private ConcurrentLinkedQueue<byte[]> outq = null;
	
	public static void main(String[] args) throws IOException {
		MbsModule mbs = new MbsModule();
		mbs.init();
	}
	public void init() throws IOException {
		qpool = new QueuePool(Param.NUM_STREAM);
		outq = new ConcurrentLinkedQueue<byte[]>();
		BroadcastThread bcast_t = new BroadcastThread(outq,Param.BSS_IP,Param.BSS_BCAST_PORT,32);
		
		for(int i=0; i<Param.NUM_STREAM;i++){
			new ReceviderThread("rcvr"+i,Param.CS_IP,Param.CS_PORT[i],Param.DATA_RATE[i],qpool.get(i)).start();
		}
		SchedulerThread sch_t = new SchedulerThread(qpool,outq);
		System.out.println("MBS ready");
		sch_t.start();
		bcast_t.start();
    }
}
class SchedulerThread extends Thread {

	private QueuePool qpool;
	private ConcurrentLinkedQueue<byte[]> outq = null;
	
	protected String name;
	public boolean keeprunning;
	Vector<byte[]>  buf;
	
	int u[];
	int f_intv[][];
	int db_intv[][];
    public SchedulerThread(QueuePool qpool,ConcurrentLinkedQueue<byte[]> outq) throws IOException {
    	super("SchedulerThread");
    	this.name = "SchedulerThread";
    
    	this.qpool = qpool;
    	this.outq = outq;
    	u =new int[Param.NUM_STREAM];
    	for(int i =0;i<Param.NUM_STREAM;i++)u[i] =0;
    	//initialize frame intervals
    	f_intv = new int[Param.NUM_FRAMES][2];//ith frame[0=start][1=end]
    	int le,re;le = 0;re = Param.FRAME_DURATION;
    	for(int i =0;i<Param.NUM_FRAMES;i++){
    		f_intv[i][0] =le;
    		f_intv[i][1] =re;
    		le = re;
    		re = re+Param.FRAME_DURATION;
    	}
    	this.keeprunning = true;
        System.out.println(this.name+" ready");
    }

    public void run() {
        while (this.keeprunning)  {
	        try {
	        	//wait for the queues to accumulate data
       		    sleep((long)(1 * 1000));
	        	int F = this.getFrameCap();// get the frame size for current mod scheme
	        	//create data block intervals
	        	db_intv = new int[Param.TDR/F][3];//ith frame[0=start][1=end][2=isbase->1]
	        	buf = new Vector<byte[]>();
	        	int le,re;
	        	for(int i =0;i<Param.NUM_STREAM;i++){
	        		byte temp[] = qpool.get(i).poll();
	        		int off = 0;
	        		le = u[i];re = le+ F*1024/Param.DATA_RATE[i];
	        		int n_blocks = Param.DATA_RATE[i]/F; 
	        		for(int j=0;j<n_blocks;j++){
	        			db_intv[i*n_blocks+j][0] = le;
	        			db_intv[i*n_blocks+j][1] = re;
	        			//System.out.println(off+" "+(F*1024));
	        			byte data[] = get_bytes(temp,off,F*1024);
	        			buf.add(data);
	        			db_intv[i*n_blocks+j][2] = Param.VALUE[i]/n_blocks;
	        			le = re; re = re +F*1024/Param.DATA_RATE[i];
	        			off = off +F*1024; 
	        		}
	        	}
	        	for(int i =0;i<db_intv.length;i++)
	        		System.out.print("["+db_intv[i][0]+","+db_intv[i][1]+"]");
	        	System.out.println();
	        	//find the matched db index of each frame
	        	//int matched_dbs[] = findMatch(db_intv,f_intv);
	        	int matched_dbs[] = new Algorithm().naiveMatch(F);
	        	
	        	//send the frames to out bound queue
	        	for(int i=0;i<matched_dbs.length;i++){
	        		System.out.print(matched_dbs[i]+" ");
	        		outq.add(buf.get(matched_dbs[i]));
	        	}
	        	System.out.println();
	        	buf = null; //drop remaining data	        	
	        	
	        }catch (Exception e) {
	        	e.printStackTrace();
	        	keeprunning = false;
	        	buf = null;
	        }
        }
    }
    /*public int[] findMatch(int data[][], int frame[][]){
//    	create node indexes
    	int source_idx =0;
    	int ystart = db_intv.length; //starting index of y nodes
    	int num_nodes = 1+ystart+f_intv.length+1;// source+ data blocks +frames +sink
    	int sink_idx = num_nodes -1;
    	
    	//create edge weight matrix.
    	int edge[][] = new int[num_nodes][num_nodes];
    	for(int i =source_idx+1;i<ystart;i++){
			for(int j = ystart; j < sink_idx;j++){
                //System.out.println("d=["+d.getS()+" "+d.getE()+"] f=["+f.getS()+" "+f.getE()+"]");
				if(f_intv[j-ystart][1] <= db_intv[i-1][1] && db_intv[i-1][0] <= f_intv[j-ystart][0]){
					//System.out.println("d=["+d.getS()+" "+d.getE()+"] f=["+f.getS()+" "+f.getE()+"]");
					edge[i][j] = db_intv[i-1][2];
				}
			}
		}
    	//create graph and find flow
    	MatrixGraph mg = new MatrixGraph(num_nodes,edge,source_idx,sink_idx);
		mg.findMaxFlow();
    	
    }*/
    public int getFrameCap(){
    	//TODO return the data rate per frame based on a modulation rate which is
    	//taken over a probability distribution and maybe by incorporating feedback
    	return 4; //in kb
    }
    public boolean isBaseLayer(int l){
    	for(int i=0;i<Param.BASE_IDX.length;i++)
    		if(l== Param.BASE_IDX[i])
    			return true;
    	return false;
    }
    
    public byte[] get_bytes(byte[] src,int off, int len){
    	byte[] temp = new byte[len];
    	for(int i =0;i<len;i++)
    		temp[i] = src[off+i];
    	return temp;
    }
    public void byte_copy(byte[] src,byte[] dest, int off){
    	for(int i =0;i<src.length;i++)
    		dest[i+off] = src[i];
    }
}
class QueuePool {
	private Vector<ConcurrentLinkedQueue<byte[]>> qpool;
	public QueuePool(int n){
		this.qpool = new Vector<ConcurrentLinkedQueue<byte[]>>(n);
		for(int i=0;i<n;i++){
			ConcurrentLinkedQueue<byte[]> clq = new ConcurrentLinkedQueue<byte[]>();
			this.qpool.add(i,clq);
		}
	}
	public ConcurrentLinkedQueue<byte[]> get(int i){ return qpool.get(i);} 
	
}

class ReceviderThread extends Thread {

    protected MulticastSocket socket = null;
    InetAddress address;
    protected boolean keepStreaming = true;
    private ConcurrentLinkedQueue<byte[]> clq = null;
    int retrieve_rate = 0;
    protected String name;
    public ReceviderThread(String name,String server_IP,int port, int r, ConcurrentLinkedQueue<byte[]> clq) throws IOException {
    	super(name);
    	this.name = name;
    	this.socket = new MulticastSocket(port);
        this.address = InetAddress.getByName(server_IP);
        socket.joinGroup(address);
        this.retrieve_rate = r;
        this.clq = clq;  
        System.out.println(name+" thread ready");
    }

    public void run() {
    	try {
    		while (keepStreaming)  {
	           	int chunk_size = this.retrieve_rate*1024; //read 8kb chunks
	        	byte[] buf = new byte[chunk_size];
	
	        	DatagramPacket packet = new DatagramPacket(buf, buf.length);
	        	this.socket.receive(packet);
	        	clq.add(buf);
	        	System.out.println(this.name+" Q received "+this.retrieve_rate+"kb data.");
	        }
    		socket.leaveGroup(this.address);
            socket.close();
    	}catch (Exception e) {
	            e.printStackTrace();
	        	keepStreaming = false;
        }
    }
}

class BroadcastThread extends Thread {

    protected MulticastSocket socket = null;
    protected boolean keepStreaming = true;
    ConcurrentLinkedQueue<byte[]> clq = null;
    int port = 0;
    String bcastIP; String name;
    int send_rate = 0;
    public BroadcastThread(ConcurrentLinkedQueue<byte[]> clq,String bcastIP,int port, int r) throws IOException {
        super("BroadcastThread");
        this.name = "BroadcastThread";
        this.bcastIP = bcastIP;
        this.port = port;
        this.send_rate = r;
        this.clq = clq;
        socket = new MulticastSocket(port);
        System.out.println("Dispather thread ready");
    }

    public void run() {
    	
        while (keepStreaming) {
	        try {
	        	int chunk_size = this.send_rate*1024; //read 8kb chunks
	        	byte[] buf = getBytes(chunk_size);
	
	        	if(buf == null || buf.length == 0)
	        		continue;
	            InetAddress group = InetAddress.getByName(this.bcastIP);
	            DatagramPacket packet = new DatagramPacket(buf, buf.length, group, this.port);
	            socket.send(packet); 
	   		    System.out.println(this.name+" sent data now.");
	       		try {
	       		    sleep((long)(1 * 1000));
	       		}catch (InterruptedException e) {
	       			e.printStackTrace();
	       		}
	        }catch (Exception e) {
	            e.printStackTrace();
	        	keepStreaming = false;
	        }
        }
        socket.close();
    }
    protected byte[] getBytes(int chunk_size) {
    	byte[] temp = null;
        try {
        	temp = new byte[chunk_size];
        	temp = clq.poll();//.readFully(temp, 0, chunk_size);
        }catch (Exception e) {
        	System.out.println(e.getMessage());e.printStackTrace();
        }
        return temp;
    }
    public void setRate(int r){
    	this.send_rate = r;
    }
}

