package src;
import java.io.File;
import java.io.IOException;
import java.io.RandomAccessFile;

public class YUVHandler
{
	public int n_frames, width, height;
	public int[][][] Y, Cb, Cr;

	public YUVHandler(String path, int n_frames, int width, int height) throws IOException{
		
		this.n_frames = n_frames;
		this.width = width;
		this.height = height;

		Y = new int[n_frames][height][width];
		Cb = new int[n_frames][height / 2][width / 2];
		Cr = new int[n_frames][height / 2][width / 2];

		RandomAccessFile fp = new RandomAccessFile(path, "r");
		byte[] temp = new byte[width * height];

		//load file contents into mem
		for(int i = 0; i < n_frames; i++)
		{
			fp.readFully(temp, 0, width * height);
			for(int j = 0; j < height; j++)
				for(int k = 0; k < width; k++)
					Y[i][j][k] = temp[j * width + k] & 0xff;

			fp.readFully(temp, 0, width * height / 4);
			for(int j = 0; j < height / 2; j++)
				for(int k = 0; k < width / 2; k++)
					Cb[i][j][k] = temp[j * width / 2 + k]  & 0xff;

			fp.readFully(temp, 0, width * height / 4);
			for(int j = 0; j < height / 2; j++)
				for(int k = 0; k < width / 2; k++)
					Cr[i][j][k] = temp[j * width / 2 + k]  & 0xff;
		}

		fp.close();
		System.out.println("% " + n_frames + " frames from YUV file " + path + " successfully read");
		//this.print();
		//System.out.println("printed");
	}

	public YUVHandler(int[][][] Y, int[][][] Cb, int[][][] Cr)
//	construct a YUV video from the given YCbCr values
	{
		n_frames = Y.length;
		height = Y[0].length;
		width = Y[0][0].length;
		this.Y = Y;
		this.Cb = Cb;
		this.Cr = Cr;
	}
	public static YUVHandler make_test(int n_frames, int width, int height)
	{
		int[][][] y = new int[n_frames][height][width];
		int[][][] cb = new int[n_frames][height / 2][width / 2];
		int[][][] cr = new int[n_frames][height / 2][width / 2];

		for(int i = 0; i < n_frames; i++)
		{
			for(int j = 0; j < height; j++)
				for(int k = 0; k < width; k++)
					y[i][j][k] = 128;

			for(int j = 0; j < height / 2; j++)
				for(int k = 0; k < width / 2; k++)
					cr[i][j][k] = cb[i][j][k] = 0;
		}

		y[n_frames - 1][1][1] = 0;
		y[n_frames - 1][2][2] = 255;

		return new YUVHandler(y, cb, cr);
	}
	public int getN(){ return n_frames;}
	public byte[] getY(int i){
		byte[] temp = new byte[width * height];
		for(int j = 0; j < height; j++)
			for(int k = 0; k < width; k++)
				temp[j * width + k] = (byte) (Y[i][j][k] & 0xff);
		return temp;
	}
	public byte[] getCb(int i){
		byte[] temp = new byte[width * height];
		for(int j = 0; j < height / 2; j++)
			for(int k = 0; k < width / 2; k++)
				temp[j * width / 2 + k] = (byte) (Cb[i][j][k] & 0xff);
		return temp;
	}
	public byte[] getCr(int i){
		byte[] temp = new byte[width * height];
		for(int j = 0; j < height / 2; j++)
			for(int k = 0; k < width / 2; k++)
				temp[j * width / 2 + k] = (byte) (Cr[i][j][k] & 0xff);
		return temp;
	}
	public void setY(int i,byte[] temp){
		for(int j = 0; j < height; j++)
			for(int k = 0; k < width; k++)
				Y[i][j][k] = temp[j * width + k] & 0xff;
	}
	public void setCb(int i,byte[] temp){	
		for(int j = 0; j < height / 2; j++)
			for(int k = 0; k < width / 2; k++)
				Cb[i][j][k] = temp[j * width / 2 + k]  & 0xff;
	}
	public void setCr(int i,byte[] temp){	
		for(int j = 0; j < height / 2; j++)
			for(int k = 0; k < width / 2; k++)
				Cr[i][j][k] = temp[j * width / 2 + k]  & 0xff;
	}
	public void export(String path) throws IOException
	{
		new File(path).delete();
		RandomAccessFile fp = new RandomAccessFile(path, "rw");
		byte[] temp = new byte[width * height];

//		read the file contents
		for(int i = 0; i < n_frames; i++)
		{
			for(int j = 0; j < height; j++)
				for(int k = 0; k < width; k++)
					temp[j * width + k] = (byte) (Y[i][j][k] & 0xff);
			fp.write(temp, 0, width * height);

			for(int j = 0; j < height / 2; j++)
				for(int k = 0; k < width / 2; k++)
					temp[j * width / 2 + k] = (byte) (Cb[i][j][k] & 0xff);
			fp.write(temp, 0, width * height / 4);

			for(int j = 0; j < height / 2; j++)
				for(int k = 0; k < width / 2; k++)
					temp[j * width / 2 + k] = (byte) (Cr[i][j][k] & 0xff);
			fp.write(temp, 0, width * height / 4);
		}

		fp.close();
	}
	
	
	
	
	public void upsample_temporal()
	{
		int[][][] new_Y =  new int[n_frames * 2][height][width];
		int[][][] new_Cb = new int[n_frames * 2][height / 2][width / 2];
		int[][][] new_Cr = new int[n_frames * 2][height / 2][width / 2];

		for(int frame = 0; frame < 2 * n_frames; frame++)
			for(int y = 0; y < height; y++)
				for(int x = 0; x < width; x++)
				{
					new_Y[frame][y][x] = Y[frame / 2][y][x];
					if(y < height / 2 && x < width / 2)
					{
						new_Cb[frame][y][x] = Cb[frame / 2][y][x];
						new_Cr[frame][y][x] = Cr[frame / 2][y][x];
					}
				}

		Y = new_Y;
		Cb = new_Cb;
		Cr = new_Cr;
		n_frames *= 2;
	}


	public void downsample_temporal()
	{
		int[][][] new_Y =  new int[n_frames / 2][height][width];
		int[][][] new_Cb = new int[n_frames / 2][height / 2][width / 2];
		int[][][] new_Cr = new int[n_frames / 2][height / 2][width / 2];

		for(int frame = 0; frame < n_frames / 2; frame++)
			for(int y = 0; y < height; y++)
				for(int x = 0; x < width; x++)
				{
					new_Y[frame][y][x] = Y[frame * 2][y][x];
					if(y < height / 2 && x < width / 2)
					{
						new_Cb[frame][y][x] = Cb[frame * 2][y][x];
						new_Cr[frame][y][x] = Cr[frame * 2][y][x];
					}
				}

		Y = new_Y;
		Cb = new_Cb;
		Cr = new_Cr;
		n_frames /= 2;
	}

	public void reduce_frame_rate(boolean[] frame_dropping_map, int frame0) throws Exception
	{
		int count = 0;
		for(int i = 0; i < frame_dropping_map.length - 1; i++)
		{
			if(frame_dropping_map[i] == true && frame_dropping_map[i + 1] == true)
				throw new Exception("frame_dropping_map cannot contain two adjacent \"true\" values");
			if(frame_dropping_map[i] == true)
				count++;
		}
		if(frame_dropping_map[frame_dropping_map.length - 1] == true)
			count++;

		for(int i = frame0; i < n_frames; i++)
		{
			if( frame_dropping_map[(i - frame0) % frame_dropping_map.length] == false )
				continue;

			for(int j = 0; j < height; j++)
				for(int k = 0; k < width; k++)
					Y[i][j][k] = Y[i - 1][j][k];

			for(int j = 0; j < height / 2; j++)
				for(int k = 0; k < width / 2; k++)
				{
					Cb[i][j][k] = Cb[i - 1][j][k];
					Cr[i][j][k] = Cr[i - 1][j][k];
				}
		}

		System.out.println( "frame rate reduced bya factor of " + count / (double) frame_dropping_map.length );
	}
	public void print(){
		for(int i = 0; i < n_frames; i++){
			for(int j = 0; j < height; j++)
				for(int k = 0; k < width; k++)
					System.out.print(""+Y[i][j][k]);

			
			for(int j = 0; j < height / 2; j++)
				for(int k = 0; k < width / 2; k++)
					System.out.print(""+Cb[i][j][k]);

			
			for(int j = 0; j < height / 2; j++)
				for(int k = 0; k < width / 2; k++)
					System.out.print(""+Cr[i][j][k]);
			
			System.out.println();
		}
	}
}
